unit hiDataToFile;


interface

uses Kol, Share, Debug;

type
  TdtProc = function (st:PStream; const Val:PData):TData of object;
  THIDataToFile = class(TDebug)
   private
   public
    _prop_Type:TdtProc;
    _prop_Data:TData;
    _prop_Count:integer;
    _prop_TrailingStr:string;
    
    _event_onGet:THI_Event;
    _event_onFound:THI_Event;
    _event_onNotFound:THI_Event;
    
    _data_Stream:THI_Event;
    _data_Dt:THI_Event;
    _data_Count:THI_Event;

    function dtByte(st: PStream; const Val: PData):TData;
    function dtWord(st: PStream; const Val: PData):TData;
    function dtCardinal(st: PStream; const Val: PData):TData;
    function dtInteger(st: PStream; const Val: PData):TData;
    function dtReal(st: PStream; const Val: PData):TData;
    function dtPString(st: PStream; const Val: PData):TData;
    function dtAnsiString(st: PStream; const Val: PData):TData;
    function dtLines(st: PStream; const Val: PData):TData;
    function dtStringLen(st: PStream; const Val: PData):TData;
    function dtLeadingString(st: PStream; const Val: PData):TData;

    procedure _work_doPut(var _Data: TData; Index: Word);
    procedure _work_doGet(var _Data: TData; Index: Word);
    procedure _work_doPosition(var _Data: TData; Index: Word);
    procedure _work_doPosEnd(var _Data: TData; Index: Word);
    procedure _work_doSkip(var _Data: TData; Index: Word);
    procedure _work_doSearch(var _Data: TData; Index: Word);
    
    procedure _var_Data(var _Data: TData; Index: Word);
    procedure _var_Position(var _Data: TData; Index: Word);
  end;
  
  function PosBuf(Buffer: Pointer; BufLen: integer; SearchBuf: Pointer; SBufLen: integer): integer;
  function SearchInStream(st: PStream; SearchBuf: Pointer; Len: integer): boolean;

implementation


//  Position            (    ). 
//    ,      2  (  Integer).
//     ,      4 .
procedure SetStreamPosition(st: PStream; NewPos: Integer);
begin
  if NewPos < 0 then
  begin
    st.Seek($7FFFFFFF, spBegin); //   2   
    st.Seek(Cardinal(NewPos)-$7FFFFFFF, spCurrent); //        
  end
  else
    st.Seek(NewPos, spBegin);
end;

//      SearchBuf   Buffer
function PosBuf(Buffer: Pointer; BufLen: integer; SearchBuf: Pointer; SBufLen: integer): integer;
var B, Tmp: PChar;
begin
  Result := -1;
  if (BufLen < 1) or (SBufLen < 1) or (BufLen < SBufLen) then exit;
  B := Buffer;
  Tmp := B + BufLen - SBufLen;
  while B <= Tmp do
  begin
    if B^ = PChar(SearchBuf)^ then
    begin
      if CompareMem(B, SearchBuf, SBufLen) then // 
      begin
        Result := B - Buffer;
        break;
      end;
    end;
    Inc(B);
  end;
end;

//     SearchBuf   st
function SearchInStream(st: PStream; SearchBuf: Pointer; Len: Integer): Boolean;
const
  BUFFER_SIZE = 2*1024*1024; MAX_SEARCH_LEN = 1*1024*1024;
  
var
  OldPos: Cardinal;
  Buffer: string;
  R, P, Need, CountInBuf: Integer;

  procedure ShiftBuffer(const FromPos: Integer);
  begin
    if FromPos > CountInBuf then
    begin
      CountInBuf := 0;
      Exit;
    end;
    CountInBuf := CountInBuf - FromPos + 1;
    Move(Buffer[FromPos], Buffer[1], CountInBuf);
  end;
  
begin
  Result := False;
  if Len < 1 then exit;

  OldPos := st.Position;

  //  Memory-   :
  if st.Memory <> nil then
  begin
    P := PosBuf(st.Memory, st.Size - OldPos, SearchBuf, Len);
    if P > -1 then // 
    begin
      //     
      st.Seek(P + Len , spCurrent);
      Result := True;
    end;
  end
  else //        
  begin
    SetLength(Buffer, BUFFER_SIZE);
    if Len > MAX_SEARCH_LEN then Len := MAX_SEARCH_LEN; //    ,     .  .

    CountInBuf := 0;

    while True do
    begin
      Need := BUFFER_SIZE - CountInBuf; //   
      R := st.Read(Buffer[CountInBuf+1], Need);
      if R > 0 then
      begin
        Inc(CountInBuf, R);
        P := PosBuf(@Buffer[1], CountInBuf, SearchBuf, Len); //     Buffer (0-based)
        if P > -1 then // 
        begin
          //     
          st.Seek(-CountInBuf + P + Len, spCurrent);
          Result := True;
          Break;
        end
        else //   ,         
          ShiftBuffer(CountInBuf - Len + 2); // (Len - 1) - ,            
      end;
      if R < Need then Break; //   
    end;

    if not Result then SetStreamPosition(st, OldPos); //    -  
  end;
end;


procedure THIDataToFile._work_doPut;
var st: PStream;
    dt: TData;
begin
   st := ToStreamEvent(_data_Stream);
   dt := ReadData(_Data, _data_Dt);
   if st <> nil then
     _prop_Type(st, @dt);
     //_prop_Type(st,@_Data);
end;

procedure THIDataToFile._var_Data;
var st: PStream;
begin
   st := ToStreamEvent(_data_Stream);
   if st <> nil then
     _Data := _prop_Type(st,nil)
   else dtNull(_Data);
end;

procedure THIDataToFile._var_Position;
var st: PStream;
begin
   st := ToStreamEvent(_data_Stream);
   if st <> nil then
     share.dtInteger(_Data,st.Position)
   else dtNull(_Data);
end;

procedure THIDataToFile._work_doPosition;
var st: PStream;
begin
   st := ToStreamEvent(_data_Stream);
   if st <> nil then
     SetStreamPosition(st, ToInteger(_Data));
end;

procedure THIDataToFile._work_doGet;
var st: PStream;
begin
   st := ToStreamEvent(_data_Stream);
   if st <> nil then
     _hi_CreateEvent(_Data,@_event_onGet,_prop_Type(st,nil));
end;

procedure THIDataToFile._work_doPosEnd;
var st: PStream;
begin
  st := ToStreamEvent(_data_Stream);
  if st <> nil then SetStreamPosition(st, st.Size);
end;

procedure THIDataToFile._work_doSkip;
var st: PStream;
begin
  st := ToStreamEvent(_data_Stream);
  if st <> nil then st.Seek(ToInteger(_Data), spCurrent);
end;

procedure THIDataToFile._work_doSearch;
var st: PStream; dt: TData;
    b: Byte; w: Word; d: Cardinal; r: Real; s: string;
    Rslt: boolean;
begin
  st := ToStreamEvent(_data_Stream);
  if st <> nil then
  begin
    dt := ReadData(_Data, _data_Dt, @_prop_Data);
    
    if @_prop_Type = @THIDataToFile.dtByte then 
    begin
      b := ToInteger(dt);
      Rslt := SearchInStream(st, @b, SizeOf(Byte));
    end else
    if @_prop_Type = @THIDataToFile.dtWord then 
    begin
      w := ToInteger(dt);
      Rslt := SearchInStream(st, @w, SizeOf(Word));
    end else
    if (@_prop_Type = @THIDataToFile.dtCardinal) or (@_prop_Type = @THIDataToFile.dtInteger) then 
    begin
      d := ToInteger(dt);
      Rslt := SearchInStream(st, @d, SizeOf(Integer));
    end else
    if @_prop_Type = @THIDataToFile.dtReal then 
    begin
      r := ToReal(dt);
      Rslt := SearchInStream(st, @r, SizeOf(Real));
    end
    else
    begin
      s := ToString(dt);
      Rslt := SearchInStream(st, @s[1], Length(s));
    end;
    
    if Rslt then _hi_CreateEvent(_Data, @_event_onFound)
    else _hi_CreateEvent(_Data, @_event_onNotFound);
  end;
end;

function THIDataToFile.dtByte;
var b: Byte;
begin
  if val = nil then
  begin
    st.read(b,1);
    Share.dtInteger(Result,b);
  end
  else
  begin
    b := ToInteger(val^);
    st.Write(b,1);
  end;
end;

function THIDataToFile.dtWord;
var w: Word;
begin
  if val = nil then
  begin
    st.read(w,2);
    Share.dtInteger(Result, w);
  end
  else
  begin
    w := ToInteger(val^);
    st.Write(w,2);
  end;
end;

function THIDataToFile.dtCardinal;
var c: Cardinal;
begin
  if val = nil then
  begin
    st.read(c,4);
    Share.dtInteger(Result,c);
  end
  else
  begin
    c := ToInteger(val^);
    st.Write(c,4);
  end;
end;

function THIDataToFile.dtInteger;
var i: Integer;
begin
  if val = nil then
  begin
    st.read(i,4);
    Share.dtInteger(Result, i);
  end
  else
  begin
    i := ToInteger(val^);
    st.Write(i,4);
  end;
end;

function THIDataToFile.dtReal;
var r: Real;
begin
  if val = nil then
  begin
    st.read(r,sizeof(real));
    Share.dtReal(Result,r);
  end
  else
  begin
    r := ToReal(val^);
    st.Write(r,sizeof(real));
  end;
end;

function THIDataToFile.dtPString;
var
  s: string;
  len: Word;
begin
  if val = nil then
  begin
    st.read(len,2);
    SetLength(s,len);
    if len > 0 then
     st.Read(s[1],len);
    Share.dtString(Result,s);
  end
  else
  begin
    s := ToString(val^);
    len := Length(s);
    st.Write(len,2);
    st.Write(s[1],len);
  end;
end;

function THIDataToFile.dtAnsiString;
begin
  if val = nil then
    Share.dtString(Result,st.ReadStrZ)
  else  st.WriteStrZ(ToString(val^));
end;

function THIDataToFile.dtLines;
begin
  if val = nil then
    Share.dtString(Result,st.ReadStr)
  else  st.WriteStr(ToString(val^)+#13#10);
end;

function THIDataToFile.dtStringLen(st:PStream; const Val:PData):TData;
var s: string;
    c, count: Integer;
begin
  count := ReadInteger(Val^, _data_Count, _prop_Count);
  if Val = nil then // 
  begin
    if count > 0 then
    begin
      SetLength(s, count); //         
      c := st.Read(s[1], count);
      if c < count then SetLength(s, c); //  ,  
    end;
    Share.dtString(Result, s);
  end
  else  // 
  begin
    if count < 1 then exit;
    s := ToString(Val^);
    c := Length(s);
    if (c < count) and (Length(_prop_TrailingStr) > 0) then //  
    begin
      SetLength(s, count);
      FillChar(s[c + 1], count - c, _prop_TrailingStr[1]);
      st.Write(s[1], Length(s));
    end
    else
      st.Write(s[1], count);
  end;
end;

function THIDataToFile.dtLeadingString(st:PStream; const Val:PData):TData;
var s: string;
    c, OldPos, k: Cardinal;
begin
  if Val = nil then // 
  begin
    c := Length(_prop_TrailingStr);
    if c > 0 then
    begin
      OldPos := st.Position; //   
      if SearchInStream(st, @_prop_TrailingStr[1], c) then
      begin
        k := st.Position - OldPos - c; //    .         
        SetLength(s, k);
        st.Position := OldPos; //  
        st.Read(s[1], k);
        st.Seek(c, spCurrent); //  
      end;
    end;
    Share.dtString(Result, s); // ,    ,   
  end
  else  // 
  begin
    st.WriteStr(ToString(ReadData(Val^, _data_Dt, @_prop_Data))+_prop_TrailingStr);
  end;
end;


end.
